"""Handles query loading, rendering, and management."""

import logging
import os
from pathlib import Path
from typing import Any, Dict, Iterator

from jinja2 import Environment, FileSystemLoader, select_autoescape

log = logging.getLogger(__name__)


def os_walk(path: Path) -> Iterator[tuple[str, list[str], list[str]]]:
    """Mimic os.walk for usage with pathlib.

    Args:
        path (Path): A Path to walk.

    Yields:
        Tuples of (root, dirs, files) equivalent to os.walk.

    """
    for root, dirs, files in os.walk(str(path)):
        yield root, dirs, files

class QueryHandler:
    """Encapsulates query management logic."""

    def __init__(self, base_directory: str) -> None:
        """Initialize QueryHandler.

        Args:
        base_directory (str): The base directory where query templates are stored.

        """
        self.queries_dir = Path(base_directory)
        self.query_dict: Dict[str, str] = {}
        #Last query string being loaded
        self.query_text: str | None = None
        self.load_queries()

    def collect_queries_from_directory(
        self, directory: Path, *, make_absolute: bool = False, base: Path | None = None,
    ) -> Dict[str, str]:
        """Collect query templates from the specified directory.

        Searches for files ending with .j2 or .graphql inside the directory.
        If make_absolute is True, returns the file paths as absolute, otherwise relative
        to the provided base (or self.queries_dir if base is None).

        Args:
            directory (Path): The directory to search.
            make_absolute (bool): Flag indicating whether to return absolute paths.
            base (Path | None): The base directory to use for relative paths.

        Returns:
            A dictionary mapping the query name (stem) to its file path as a string.

        """
        collected: Dict[str, str] = {}
        for root, _, files in os_walk(directory):
            for file_name in files:
                if file_name.endswith((".j2", ".graphql")):
                    query_name = Path(file_name).stem
                    file_path = (Path(root) / file_name).resolve()
                    if make_absolute:
                        collected[query_name] = str(file_path)
                    else:
                        base_dir = base if base is not None else self.queries_dir
                        relative_path = file_path.relative_to(base_dir)
                        collected[query_name] = str(relative_path)
        return collected

    def load_queries(self) -> Dict[str, str]:
        """Load query templates in a dictionary from the base queries directory.

        Allow to load the default queries from the queries directory.

        Returns:
            A dictionary mapping query names to their paths.

        """
        self.query_dict = self.collect_queries_from_directory(
            self.queries_dir,
            make_absolute=False,
        )
        return self.query_dict

    def add_queries_from_directory(self, directory: str) -> None:
        """Add all query templates from a new directory.

        Searches for query files in the directory and adds them
        to the query dictionary with their absolute paths.

        Args:
            directory (str): The path to the directory that contains query templates.

        """
        new_dir = Path(directory).resolve()
        if not new_dir.is_dir():
            msg = f"Provided directory {directory} is not a valid directory."
            raise ValueError(msg)
        new_queries = self.collect_queries_from_directory(new_dir, make_absolute=True)
        for query_name, file_path in new_queries.items():
            if query_name in self.query_dict:
                log.warning("Query '%s' \
                already exists and will be overridden.", query_name)
            self.query_dict[query_name] = file_path
        log.info("Queries loaded from directory: %s", new_dir)

    def load_query(self, query_name: str,
                   jinja_variables: dict[str, Any] | None,
                   ) -> str:
        """Load and render a query template by name.

        Args:
            query_name (str): The name (stem) of the query file to load.
            jinja_variables (dict[str, Any] | None): To render into a Jinja template.

        Returns:
            The loaded query text.

        Raises:
            ValueError: If the query file does not exist or is unsupported.

        """
        template_path = self.query_dict.get(query_name)
        if not template_path:
            msg = f"Query '{query_name}' not found in query_dict."
            raise ValueError(msg)

        full_path = Path(template_path) if Path(template_path).is_absolute() \
            else self.queries_dir / template_path

        if template_path.endswith(".j2"):
            jinja_env = Environment(
                loader=FileSystemLoader(str(full_path.parent)),
                autoescape=select_autoescape(),
                trim_blocks=True,
                lstrip_blocks=True,
            )
            template = jinja_env.get_template(full_path.name)
            query_text = template.render(**(jinja_variables or {}))
        elif template_path.endswith(".graphql"):
            with full_path.open("r", encoding="utf-8") as file:
                query_text = file.read()
        else:
            msg = f"Unsupported query file type: '{template_path}'."
            raise ValueError(msg)

        self.query_text = query_text
        return query_text

    def add_query(self, query_name: str, query_path: str) -> None:
        """Add a new query to the list of queries.

        Args:
            query_name (str): The name (stem) of the query.
            query_path (str): The relative or absolute path to the query template.

        """
        if query_name in self.query_dict:
            log.warning("Query '%s' already exists and will be overridden.", query_name)
        self.query_dict[query_name] = query_path

    def delete_queries(self, queries_to_delete: list[str]) -> None:
        """Delete specified queries from the query dictionary.

        Args:
            queries_to_delete (List[str]): A list of query names to delete.

        """
        for query_name in queries_to_delete:
            if query_name in self.query_dict:
                del self.query_dict[query_name]
                log.info("Deleted query: %s", query_name)
            else:
                log.warning("Query '%s' not found in query_dict. \
                 No action taken.", query_name)



"""Conftest for Piscada Foresight tests, providing shared fixtures."""

from typing import Any, Dict
from unittest.mock import Mock

import pytest

from piscada_foresight.parser_graphql_to_df. \
    time_series.timeseries_response_parser import (
    TimeseriesResponseParser,
)
from tests.utils.test_utils import load_file_content


@pytest.fixture
def parser() -> TimeseriesResponseParser:
    """Return an instance of TimeseriesResponseParser."""
    return TimeseriesResponseParser()


@pytest.fixture
def mock_query_manager() -> Mock:
    """Return a mock implementation of the QueryManager."""
    return Mock()


@pytest.fixture
def aggregated_timeseries_response() -> Dict[str, Any]:
    """Return the aggregated timeseries response from JSON."""
    return load_file_content("responses/aggregated_timeseries.json")


@pytest.fixture
def aggregated_one_empty_sublist_response() -> Dict[str, Any]:
    """Return aggregated data with one empty sublist from JSON."""
    return load_file_content("responses/aggregated_one_empty_sublist.json")


@pytest.fixture
def aggregated_tag_missing() -> Dict[str, Any]:
    """Return aggregated data missing a tag from JSON."""
    return load_file_content("responses/aggregated_tag_missing.json")


@pytest.fixture
def aggregated_empty_sublists() -> Dict[str, Any]:
    """Return aggregated data with empty sublists from JSON."""
    return load_file_content("responses/aggregated_empty_sublists.json")


@pytest.fixture
def aggregated_no_values() -> Dict[str, Any]:
    """Return aggregated data with no values from JSON."""
    return load_file_content("responses/aggregated_no_values.json")


@pytest.fixture
def raw_missing_values_tag_response() -> Dict[str, Any]:
    """Return raw data missing values for some tags from JSON."""
    return load_file_content("responses/raw_missing_values_tag.json")


@pytest.fixture
def raw_timeseries_response() -> Dict[str, Any]:
    """Return raw timeseries response from JSON."""
    return load_file_content("responses/raw_timeseries.json")


@pytest.fixture
def no_values_response(raw_timeseries_response: Dict[str, Any]) -> Dict[str, Any]:
    """Return a raw timeseries response with all values removed."""
    response = raw_timeseries_response
    response["entityId_0"]["trait"]["quantity"]["values"] = []
    response["entityId_1"]["trait"]["quantity"]["values"] = []
    return response



"""Parse GraphQL timeseries responses into Pandas DataFrames."""

import logging
from datetime import datetime
from typing import Any

from pandas import DataFrame, concat, to_datetime

log = logging.getLogger(__name__)


class TimeseriesResponseParser:
    """Converting GraphQL timeseries to Pandas DataFrames.

    Methods
    -------
    parse(
        response: dict,
        entity_variables: dict,
        start: datetime,
        end: datetime,
        query_type: str
    ) -> DataFrame
        High-level entry point. Decides which parser to call based on `query_type`.

    _parse_raw_values(
        response: dict,
        entity_variables: dict,
        start: datetime,
        end: datetime
    ) -> DataFrame
        Parses response from a 'get_raw_values' GraphQL query.

    _parse_aggregated_values(
        response: dict,
        entity_variables: dict,
        start: datetime,
        end: datetime
    ) -> DataFrame
        Parses response from a 'get_aggregated_values' GraphQL query.

    """

    def parse(
            self,
            response: dict[str, Any],
            entity_variables: dict[str, str],
            start: datetime,
            end: datetime,
            query_type: str="raw",
            ) -> DataFrame:
        """Dispatch to the appropriate parser method based on `query_type`.

        Parameters
        ----------
        response : dict
            The raw GraphQL response (already converted to Python dict).
        entity_variables : dict
            A mapping like {"entityId_0": "some_entity_id", ...}.
        start : datetime
            The start time used in the query (for error messages).
        end : datetime
            The end time used in the query (for error messages).
        query_type : str
            Determines which parser to use, e.g. "raw" or "aggregated".

        Returns
        -------
        DataFrame
            The parsed DataFrame.

        """
        if query_type == "raw":
            return self._parse_raw_values(response, entity_variables, start, end)
        if query_type == "aggregated":
            return self._parse_aggregated_values(response, entity_variables, start, end)
        msg = f"Unknown query_type: {query_type}"
        raise ValueError(msg)

    def _parse_raw_values(self,
                          response: dict[str, Any],
                          entity_variables: dict[str, str],
                          start: datetime,
                          end: datetime,
                          ) -> DataFrame:
        """Parse a 'get_raw_values' style response.

        response = {
          "entityId_0": {
            "name": "...",
            "trait": {
              "quantity": {
                "values": [
                  { "eventTime": "...", "value": "..." },
                  ...
                ]
              }
            }
          },
          ...
        }.
        """
        series_list = []

        for variable_name, entity_id in entity_variables.items():
            try:
                entity_data = response[variable_name]
                name = entity_data["name"]
                values = entity_data["trait"]["quantity"]["values"]

                # Convert to Series
                frame = DataFrame(values).set_index("eventTime")["value"]
                frame.index = to_datetime(frame.index, format="ISO8601")
                frame = frame.astype(float)
                frame.name = name
                series_list.append(frame)

            except KeyError as exc:
                msg = (
                    f"Could not retrieve raw values for entity '{entity_id}' "
                    f"in time range {start} - {end}. Missing key: {exc}"
                )
                log.exception(msg)
                raise RuntimeError(msg) from exc
            except TypeError as exc:
                msg = f"Could not find entity '{entity_id}'."
                log.exception(msg)
                raise RuntimeError(msg) from exc

        if not series_list:
            return DataFrame()

        # Merge all entities' series side by side
        df_merged = concat(series_list, axis=1)
        return df_merged.ffill()  # optional forward-fill

    def _parse_aggregated_values(
                                self,
                                response: dict[str, Any],
                                entity_variables: dict[str, str],
                                start: datetime,
                                end: datetime,
                                ) -> DataFrame:
        """Parse a 'get_aggregated_values' style response.

        response = {
          "entityId_0": {
            "name": "...",
            "trait": {
              "quantity": {
                "aggregatedTimeseries": [
                  [{"eventTime":".","value":".","aggregationFunction":"min" },],
                  [{"eventTime":".","value":".","aggregationFunction":"max" },],
                  ...
                ]
              }
            }
          },
          ...
        }.
        """
        all_frames = []

        for variable_name, entity_id in entity_variables.items():
            try:
                entity_data = response[variable_name]
                name = entity_data["name"]
                agg_series_list = entity_data["trait"]["quantity"][
                    "aggregatedTimeseries"
                ]
            except KeyError as exc:
                msg = (
                    f"Could not retrieve aggregated values for entity '{entity_id}' "
                    f"in time range {start} - {end}. Missing key: {exc}"
                )
                log.exception(msg)
                raise RuntimeError(msg) from exc

            frames_for_this_entity = []
            for sublist in agg_series_list:
                if not sublist:
                    continue
                tmp_df = DataFrame(sublist).set_index("eventTime")
                tmp_df.index = to_datetime(tmp_df.index, format="ISO8601")

                # aggregator name is typically the same for all rows
                aggregator = tmp_df["aggregationFunction"].iloc[0]
                col_name = f"{name}|{aggregator}"

                tmp_df["value"] = tmp_df["value"].astype(float)
                tmp_df = tmp_df.rename(columns={"value": col_name})

                # keep only the aggregator column
                frames_for_this_entity.append(tmp_df[[col_name]])

            if frames_for_this_entity:
                entity_df = concat(frames_for_this_entity, axis=1)
                all_frames.append(entity_df)

        if not all_frames:
            log.warning(
                "No data found for the requested time range \
                %s - %s. Returning an empty DataFrame.", start, end,
            )
            return DataFrame()

        # Merge horizontally across all entities
        merged_dataframe = concat(all_frames, axis=1)
        return merged_dataframe.ffill()



"""Access to Foresight timeseries data."""

import logging
import re
from datetime import datetime, timezone
from json import loads
from uuid import UUID

from pandas import DataFrame, to_datetime

from piscada_foresight.parser_graphql_to_df. \
    time_series.timeseries_response_parser import TimeseriesResponseParser
from piscada_foresight.queries_templates.query_manager import QueryManager

log = logging.getLogger(__name__)


def get_value(
    query_manager: QueryManager, entity_id: UUID, moment: datetime | None = None,
) -> float:
    """Retrieve the latest value of a `foresight:Datapoint` entity."""
    if not moment:
        moment = datetime.now(tz=timezone.utc)

    variables = {"entityId": entity_id, "eventTime": moment.isoformat()}
    response = query_manager.run_query("get_latest_value", query_variables=variables)

    try:
        return float(response["entity"]["trait"]["quantity"]["value"]["value"])
    except KeyError as err:
        msg = "Cloud not retrieve value."
        raise RuntimeError(msg) from err
    except ValueError as exc:
        val_str = response["entity"]["trait"]["quantity"]["value"]["value"]
        msg = f"Could not parse value {val_str}"
        raise RuntimeError(msg) from exc


def get_raw_timeseries(
    query_manager: QueryManager,
    entity_ids: list[str],
    start: datetime,
    end: datetime | None = None,
) -> DataFrame:
    """Retrieve values of `foresight:Datapoint` entities for a time range."""
    if end is None:
        end = datetime.now(tz=timezone.utc)
    if start.tzinfo is None or start.tzinfo.utcoffset(start) is None:
        msg = "The start parameter must be timezone aware."
        raise ValueError(msg)
    if end.tzinfo is None or end.tzinfo.utcoffset(end) is None:
        msg = "The end parameter must be timezone aware."
        raise ValueError(msg)
    if start > end:
        msg = "The 'start' datetime cannot be later than the 'end' datetime."
        raise ValueError(
            msg,
        )

    entity_variables = {
        f"entityId_{i}": entity_id for i, entity_id in enumerate(entity_ids)
    }
    jinja_dict = {}
    jinja_dict["variable_names"] = list(entity_variables.keys())
    variables = {
        **entity_variables,
        "startEventTime": start.isoformat(),
        "endEventTime": end.isoformat(),
    }

    response = query_manager.run_query(
        "get_raw_values",
        jinja_variables=jinja_dict,
        query_variables=variables,
    )

    parser = TimeseriesResponseParser()
    return parser.parse(response, entity_variables, start, end, query_type="raw")


def get_aggregated_timeseries(  # noqa: PLR0913
    query_manager: QueryManager,
    entity_ids: list[str],
    start: datetime,
    end: datetime | None = None,
    interval: str = "",
    aggregation_functions: list[str] | None = None,
) -> DataFrame:
    """Retrieve agg values of `foresight:Datapoint` entities for a given time range."""
    if end is None:
        end = datetime.now(tz=timezone.utc)
    if start.tzinfo is None or start.tzinfo.utcoffset(start) is None:
        msg = "The start parameter must be timezone aware."
        raise ValueError(msg)
    if end.tzinfo is None or end.tzinfo.utcoffset(end) is None:
        msg = "The end parameter must be timezone aware."
        raise ValueError(msg)
    if start > end:
        msg = "The 'start' datetime cannot be later than the 'end' datetime."
        raise ValueError(
            msg,
        )

    if aggregation_functions is None:
        aggregation_functions = []

    # The values are "most likely" cached/saved. (1y is the max interval)
    if interval == "":
        interval = "1y"

    valid_aggregators = {"min", "max", "avg", "count", "last"}
    if not all(func in valid_aggregators for func in aggregation_functions):
        msg = (
            f"Invalid aggregation function provided. "
            f"Only the following are allowed: {valid_aggregators}"
        )
        raise ValueError(
            msg,
        )

    entity_variables = {
        f"entityId_{i}": entity_id for i, entity_id in enumerate(entity_ids)
    }

    query_variables = {
        **entity_variables,
        "startEventTime": start.isoformat(),
        "endEventTime": end.isoformat(),
        "interval": interval,
        "aggregationFunctions": aggregation_functions,
    }

    jinja_dict = {
        "variable_names": list(entity_variables.keys()),
    }
    response = query_manager.run_query(
        "get_aggregated_values",
        jinja_variables=jinja_dict,
        query_variables=query_variables,
    )
    parser = TimeseriesResponseParser()
    return parser.parse(response, entity_variables, start, end, query_type="aggregated")


def get_values(  # noqa: PLR0913
    query_manager: QueryManager,
    entity_ids: list[str],
    start: datetime,
    end: datetime | None = None,
    interval: str = "",
    aggregation_functions: list[str] | None = None,
) -> DataFrame:
    """Dispatches either to raw or aggregated timeseries."""
    if not aggregation_functions:
        return get_raw_timeseries(query_manager, entity_ids, start, end)
    return get_aggregated_timeseries(
        query_manager, entity_ids, start, end, interval, aggregation_functions,
    )


def get_all_values(text: str) -> list[DataFrame]:
    """Extract all pairs of id, name, and values from a GraphQL query."""
    values_regex = re.compile(
        r"'id'\:\s*'([\:\w\s\-_]*?)',\s*'name'\:\s*'([\w\s\-_]*?)',\s*'trait':\s*{'quantity'\:\s*\{'values'\:\s*\[([\{'\w\:\-\s\.\},]*)",
    )
    dfs = []
    for match in values_regex.findall(text):
        eid = match[0].split(":")[-1]
        name = match[1]
        # Convert single quotes to double quotes for JSON parsing
        values_str = f"[{match[2]}]".replace("'", '"')

        parsed_values = loads(values_str)
        data = [
            {"ts": item["eventTime"], f"{name}|{eid}": float(item["value"])}
            for item in parsed_values
        ]
        df_values = DataFrame(data).set_index("ts")
        df_values.index = to_datetime(df_values.index, format="ISO8601")
        dfs.append(df_values)

    return dfs



"""Utility functions for file and DataFrame operations used in tests."""
import json
from pathlib import Path
from typing import Union, cast

import pandas as pd


def load_file_content(filename: str) -> Union[dict, str]:
    """Load content from a text, JSON, or GraphQL file.

    The file is expected to be located in the test_config directory.

    Args:
        filename (str): The name of the file to load, including its extension.

    Returns:
        dict | str: Parsed JSON data if the file is a JSON file, or raw string
            content for other files.

    Raises:
        FileNotFoundError: If the file does not exist.
        ValueError: If the file extension is unsupported.

    """
    config_path = Path(__file__).parent / "../test_config" / filename
    if not config_path.exists():
        msg = f"File not found: {config_path.resolve()}"
        raise FileNotFoundError(msg)

    file_extension = config_path.suffix.lower()
    with config_path.open("r", encoding="utf-8") as file:
        if file_extension == ".json":
            return json.load(file)
        if file_extension in {".txt", ".graphql"}:
            return file.read()
        msg = f"Unsupported file type: {file_extension}"
        raise ValueError(msg)


def save_dataframe(dataframe: pd.DataFrame, path: str) -> None:
    """Save a DataFrame to a CSV file with microsecond precision.

    Args:
        dataframe (pd.DataFrame): The DataFrame to save.
        path (str): The file path where the DataFrame will be saved.

    Returns:
        None

    """
    save_path = Path(path)
    save_path.parent.mkdir(parents=True, exist_ok=True)
    dataframe.to_csv(save_path, index=True, date_format="%Y-%m-%dT%H:%M:%S.%f%z")


def load_dataframe(path: str) -> pd.DataFrame:
    """Load a DataFrame from a CSV file and ensure its index is timezone-aware (UTC).

    Args:
        path (str): The file path from which the CSV file is loaded.

    Returns:
        pd.DataFrame: The loaded DataFrame with a timezone-aware index.

    Raises:
        FileNotFoundError: If the CSV file does not exist.
        TypeError: If the loaded index is not a DatetimeIndex.

    """
    save_path = Path(path)
    if not save_path.exists():
        msg = f"File not found: {save_path}"
        raise FileNotFoundError(msg)

    dataframe = pd.read_csv(save_path, index_col=0, parse_dates=True)
    if not isinstance(dataframe.index, pd.DatetimeIndex):
        msg = "Index is not a DatetimeIndex"
        raise TypeError(msg)
    dataframe.index = cast(pd.DatetimeIndex, dataframe.index)
    if dataframe.index.tz is None:
        dataframe.index = dataframe.index.tz_localize("UTC")
    else:
        dataframe.index = dataframe.index.tz_convert("UTC")

    return dataframe


def compare_dataframes(df1: pd.DataFrame, df2: pd.DataFrame) -> None:
    """Compare two DataFrames by flooring their indexes to the nearest second.

    Args:
        df1 (pd.DataFrame): The first DataFrame to compare.
        df2 (pd.DataFrame): The second DataFrame to compare.

    Raises:
        AssertionError: If the DataFrames are not equal.

    """
    if not isinstance(df1.index, pd.DatetimeIndex):
        msg = "df1.index is not a DatetimeIndex"
        raise TypeError(msg)
    df1.index = cast(pd.DatetimeIndex, df1.index)

    if not isinstance(df2.index, pd.DatetimeIndex):
        msg = "df2.index is not a DatetimeIndex"
        raise TypeError(msg)
    df2.index = cast(pd.DatetimeIndex, df2.index)

    df1.index = df1.index.floor("S")
    df2.index = df2.index.floor("S")

    pd.testing.assert_frame_equal(
        df1,
        df2,
        check_dtype=False,
        check_exact=False,
    )



"""Tests for the piscada_foresight.data module, verifying get_values function behaviors.

This module includes tests for successful data retrieval, error conditions, and
aggregated data handling.
"""

from datetime import datetime, timezone
from typing import Any, Dict
from unittest.mock import MagicMock

import pandas as pd
import pytest

from piscada_foresight.data import get_values
from piscada_foresight.queries_templates.query_manager import QueryManager


def test_get_values_success(
        mock_query_manager: MagicMock,
        raw_timeseries_response: Dict[str, Any],
) -> None:
    """Tests that get_values successfully returns a DataFrame with expected data."""
    mock_query_manager.run_query.return_value = raw_timeseries_response
    data_frame = get_values(
        mock_query_manager,
        entity_ids=[
            "brick:Supply_Air_Temperature_Sensor:00000000-0000-0000-0000-000000000001",
            "brick:Effective_Supply_Air_Temperature_Setpoint:00000000-0000-0000-0000-000000000002",
        ],
        start=datetime.now(tz=timezone.utc),
    )
    expected_value = 8
    if len(data_frame) != expected_value:
        pytest.fail("Expected 8 rows in the DataFrame.")
    if not pd.isna(data_frame["360001 Temperature"].iloc[0]):
        pytest.fail(
            "Expected NaN in the first row of '360001 Temperature' column.",
        )


def test_get_values_missing_values(
        mock_query_manager: MagicMock,
        no_values_response: Dict[str, Any],
) -> None:
    """Tests that get_values raises a RuntimeError when no data is available."""
    mock_query_manager.run_query.return_value = no_values_response
    with pytest.raises(RuntimeError):
        get_values(
            mock_query_manager,
            entity_ids=[
                "brick:Supply_Air_Temperature_Sensor:00000000-0000-0000-0000-000000000001",
                "brick:Effective_Supply_Air_Temperature_Setpoint:00000000-0000-0000-0000-000000000002",
            ],
            start=datetime(1993, 2, 12, tzinfo=timezone.utc),
            end=datetime(1993, 2, 13, tzinfo=timezone.utc),
        )


def test_get_values_start_later_than_end(mock_query_manager: MagicMock) -> None:
    """Tests that get_values raises a ValueError when start > end."""
    with pytest.raises(
            ValueError,
            match="The 'start' datetime cannot be later than the 'end' datetime.",
    ):
        get_values(
            query_manager=mock_query_manager,
            entity_ids=["entity1", "entity2"],
            start=datetime(2023, 2, 1, tzinfo=timezone.utc),
            end=datetime(2023, 1, 31, tzinfo=timezone.utc),
        )


def test_get_values_start_or_end_not_timezone_aware(
        mock_query_manager: MagicMock,
) -> None:
    """Tests get_values raises a ValueError if start or end is not tz-aware."""
    # Separate raises for PT012 compliance:
    with pytest.raises(ValueError, match="The start parameter must be timezone aware."):
        get_values(
            query_manager=mock_query_manager,
            entity_ids=["entity1", "entity2"],
            start=datetime(2023, 2, 1),  # no tzinfo  # noqa: DTZ001
            end=datetime(2023, 1, 31, tzinfo=timezone.utc),
        )

    with pytest.raises(ValueError, match="The start parameter must be timezone aware."):
        get_values(
            query_manager=mock_query_manager,
            entity_ids=["entity1", "entity2"],
            start=datetime(2023, 2, 1),  # no tzinfo  # noqa: DTZ001
            end=datetime(2023, 1, 31),  # no tzinfo  # noqa: DTZ001
        )

    with pytest.raises(ValueError, match="The end parameter must be timezone aware."):
        get_values(
            query_manager=mock_query_manager,
            entity_ids=["entity1", "entity2"],
            start=datetime(2023, 2, 1, tzinfo=timezone.utc),
            end=datetime(2023, 1, 31),  # no tzinfo  # noqa: DTZ001
        )


def test_get_values_aggregated_success(
        mock_query_manager: MagicMock,
        aggregated_timeseries_response: Dict[str, Any],
) -> None:
    """Checks that get_values retrieves aggregated data when aggregators are provided.

    Ensures that columns for each aggregator function exist in the result.
    """
    mock_query_manager.run_query.return_value = aggregated_timeseries_response
    data_frame = get_values(
        query_manager=mock_query_manager,
        entity_ids=[
            "brick:Supply_Air_Temperature_Sensor:00000000-0000-0000-0000-000000000001",
            "brick:Effective_Supply_Air_Temperature_Setpoint:00000000-0000-0000-0000-000000000002",
        ],
        start=datetime(2025, 1, 14, 9, 0, 0, tzinfo=timezone.utc),
        end=datetime(2025, 1, 14, 11, 0, 0, tzinfo=timezone.utc),
        interval="1h",
        aggregation_functions=["min", "max", "avg", "count", "last"],
    )
    if data_frame.empty:
        pytest.fail("DataFrame should not be empty for valid aggregated data.")

    expected_columns = [
        "360001 Temperature|min",
        "360001 Temperature|max",
        "360001 Temperature|avg",
        "360001 Temperature|count",
        "360001 Temperature|last",
    ]
    for col in expected_columns:
        if col not in data_frame.columns:
            pytest.fail(f"Missing expected column: {col}")


def test_get_values_aggregated_invalid_aggregator(mock_query_manager: MagicMock)\
        -> None:
    """Verifies that ValueError is raised if any aggregator is invalid."""
    with pytest.raises(ValueError, match="Invalid aggregation function provided"):
        get_values(
            query_manager=mock_query_manager,
            entity_ids=["entityId_0"],
            start=datetime(2025, 1, 14, tzinfo=timezone.utc),
            end=datetime(2025, 1, 15, tzinfo=timezone.utc),
            interval="1d",
            aggregation_functions=["min", "max", "foo"],  # 'foo' is invalid
        )


def test_get_values_aggregated_multiple_aggregators(
        mock_query_manager: MagicMock,
        aggregated_timeseries_response: Dict[str, Any],
) -> None:
    """Ensures that multiple valid aggregation functions are handled correctly."""
    mock_query_manager.run_query.return_value = aggregated_timeseries_response
    data_frame = get_values(
        query_manager=mock_query_manager,
        entity_ids=["entityId_0", "entityId_1"],
        start=datetime(2025, 1, 14, tzinfo=timezone.utc),
        end=datetime(2025, 1, 15, tzinfo=timezone.utc),
        interval="1h",
        aggregation_functions=["min", "max", "avg"],
    )
    if data_frame.empty:
        pytest.fail("Aggregated DataFrame should not be empty.")

    for func in ["min", "max", "avg"]:
        col_0 = f"360001 Temperature|{func}"
        col_1 = f"360002 Temperature Setpoint|{func}"
        if col_0 not in data_frame.columns:
            pytest.fail(f"Missing {col_0} aggregator column.")
        if col_1 not in data_frame.columns:
            pytest.fail(f"Missing {col_1} aggregator column.")


def test_get_values_aggregated_interval_none_uses_default(
        mock_query_manager: MagicMock,
        aggregated_timeseries_response: Dict[str, Any],
) -> None:
    """If interval=None is passed, get_values should default to an interval (e.g. '1d').

    Confirms no error is raised and the query can run successfully.
    """
    mock_query_manager.run_query.return_value = aggregated_timeseries_response
    data_frame = get_values(
        query_manager=mock_query_manager,
        entity_ids=["entityId_0"],
        start=datetime(2025, 1, 14, tzinfo=timezone.utc),
        end=datetime(2025, 1, 15, tzinfo=timezone.utc),
        interval=None,
        aggregation_functions=["min", "max", "avg", "count", "last"],
    )
    if data_frame.empty:
        pytest.fail("DataFrame should not be empty with default interval.")


def test_get_values_aggregated_start_later_than_end_raises_error(
        mock_query_manager: MagicMock,
) -> None:
    """Ensures ValueError is raised when start > end for aggregated queries."""
    with pytest.raises(
            ValueError,
            match="The 'start' datetime cannot be later than the 'end' datetime.",
    ):
        get_values(
            query_manager=mock_query_manager,
            entity_ids=["entityId_0"],
            start=datetime(2025, 1, 15, tzinfo=timezone.utc),
            end=datetime(2025, 1, 14, tzinfo=timezone.utc),
            aggregation_functions=["min"],
        )


def test_get_values_aggregated_start_or_end_not_timezone_aware_raises_error(
    mock_query_manager: MagicMock,
) -> None:
    """Checks that ValueError is raised if start or end is not tz aware."""
    with pytest.raises(ValueError, match="The start parameter must be timezone aware."):
        get_values(
            query_manager=mock_query_manager,
            entity_ids=["entityId_0"],
            start=datetime(2025, 1, 14),  # no tzinfo  # noqa: DTZ001
            end=datetime(2025, 1, 15, tzinfo=timezone.utc),
            aggregation_functions=["min"],
        )

    with pytest.raises(ValueError, match="The end parameter must be timezone aware."):
        get_values(
            query_manager=mock_query_manager,
            entity_ids=["entityId_0"],
            start=datetime(2025, 1, 14, tzinfo=timezone.utc),
            end=datetime(2025, 1, 15),  # no tzinfo  # noqa: DTZ001
            aggregation_functions=["min"],
        )

@pytest.mark.skip(reason="Need the right authorization to test the API")
def test_get_values_api() -> None:
    """Tests get_values by calling the real API (skipped unless authorized)."""
    domain = "foresight.piscada.cloud"
    query_manager = QueryManager(domain=domain,
                                 client_id="foresight-lib-py",
                                 client_secret=None,
                                 graphql_endpoint=None,
                                 json_serialize=None)
    result = get_values(
        query_manager=query_manager,
        entity_ids=[
            "brick:Supply_Air_Temperature_Sensor:0192576a-f715-72a9-826b-aa4d1c37882b",
        ],
        start=datetime(2025, 1, 14, tzinfo=timezone.utc),
        end=datetime(2025, 1, 15, tzinfo=timezone.utc),
        aggregation_functions=["min", "max", "avg", "count", "last"],
    )
    if not isinstance(result, pd.DataFrame):
        pytest.fail("The result should be a DataFrame.")
    if len(result) == 0:
        pytest.fail("The DataFrame should not be empty.")

    domain = "foresight.piscada.cloud"
    query_manager = QueryManager(domain=domain,
                                 client_id="foresight-lib-py",
                                 client_secret=None,
                                 graphql_endpoint=None,
                                 json_serialize=None)
    result = get_values(
        query_manager=query_manager,
        entity_ids=[
            "brick:Supply_Air_Temperature_Sensor:0192576a-f715-72a9-826b-aa4d1c37882b",
        ],
        start=datetime(2025, 1, 14, tzinfo=timezone.utc),
        end=datetime(2025, 1, 15, tzinfo=timezone.utc),
    )
    if not isinstance(result, pd.DataFrame):
        pytest.fail("The result should be a DataFrame.")
    if len(result) == 0:
        pytest.fail("The DataFrame should not be empty.")


@pytest.mark.skip(reason="Need the right authorization to test the API")
def test_get_values_api_wrong_entity_id() -> None:
    """Get_values with wrong entity ID."""
    domain = "foresight.piscada.cloud"
    query_manager = QueryManager(domain=domain,
                                 client_id="foresight-lib-py",
                                 client_secret=None,
                                 graphql_endpoint=None,
                                 json_serialize=None)
    with pytest.raises(RuntimeError):
        get_values(
            query_manager=query_manager,
            entity_ids=["wrong_id"],
            start=datetime(2025, 1, 14, tzinfo=timezone.utc),
            end=datetime(2025, 1, 15, tzinfo=timezone.utc),
        )



"""Test suite for the domains module of piscada foresight.

This module includes tests for domain validation, retrieving domains, traits,
and relationships.
"""

import json
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from piscada_foresight.domains import (
    _validate_domains,
    get_domains,
    get_parent_traits,
    get_trait_by_id,
)
from piscada_foresight.model import Trait

# Load the JSON response used in tests
DOMAINS_RESPONSE = json.load((Path(__file__).parent / "domains_response.json").open())


def test_get_domains_count(mocker: MockerFixture) -> None:
    """Test that the correct number of domains is retrieved from the mock client."""
    mock_client = mocker.Mock()
    mock_client.run_query.return_value = DOMAINS_RESPONSE["data"]
    domains = get_domains(mock_client)

    if len(domains) != 16:  # noqa: PLR2004
        pytest.fail("Expected 16 domains, found a different number.")


def test_get_domains_brick_domain(mocker: MockerFixture) -> None:
    """Test that the Brick domain has the correct name, prefix, and URI."""
    mock_client = mocker.Mock()
    mock_client.run_query.return_value = DOMAINS_RESPONSE["data"]
    domains = get_domains(mock_client)

    brick_domain = domains[2]
    if brick_domain.name != "Brick":
        pytest.fail('Expected the third domain to have name "Brick".')
    if brick_domain.prefix != "brick":
        pytest.fail('Expected prefix to be "brick".')
    if brick_domain.uri != "https://brickschema.org/schema/Brick#":
        pytest.fail('Expected URI to be "https://brickschema.org/schema/Brick#".')


def test_get_domains_brick_traits(mocker: MockerFixture) -> None:
    """Test that specific traits within the Brick domain match expected values."""
    mock_client = mocker.Mock()
    mock_client.run_query.return_value = DOMAINS_RESPONSE["data"]
    domains = get_domains(mock_client)

    brick_domain = domains[2]
    # Check Access_Control_Equipment trait
    ace_trait = brick_domain.traits[2]
    if ace_trait.id != "brick:Access_Control_Equipment":
        pytest.fail('Expected trait ID to be "brick:Access_Control_Equipment".')
    if ace_trait.parent_ids != ["brick:Security_Equipment"]:
        pytest.fail('Expected parent IDs to be ["brick:Security_Equipment"].')
    if ace_trait.child_ids != ["brick:Access_Reader"]:
        pytest.fail('Expected child IDs to be ["brick:Access_Reader"].')

    # Check AHU trait
    ahu_trait = brick_domain.traits[6]
    if ahu_trait.id != "brick:AHU":
        pytest.fail('Expected trait ID to be "brick:AHU".')
    expected_equiv = ["brick:Air_Handler_Unit", "brick:Air_Handling_Unit"]
    if ahu_trait.equivalent_ids != expected_equiv:
        pytest.fail(f"Expected equivalent IDs to be {expected_equiv}.")


def test_get_domains_brick_relationships(mocker: MockerFixture) -> None:
    """Test that relationships within the Brick domain match expected values."""
    mock_client = mocker.Mock()
    mock_client.run_query.return_value = DOMAINS_RESPONSE["data"]
    domains = get_domains(mock_client)

    brick_domain = domains[2]
    feeds_rel = brick_domain.relationships[0]
    if feeds_rel.id != "brick:feeds":
        pytest.fail('Expected relationship ID to be "brick:feeds".')

    expected_parents = [
        "fs_navigation:breadcrumbProcessIncoming",
        "fs_navigation:Expert",
        "fs_navigation:moreDetailsOrOutgoing",
        "fs_navigation:outgoing",
        "fs_navigation:Semantic",
        "fs_schema:Association",
    ]
    if feeds_rel.parent_ids != expected_parents:
        pytest.fail(f"Expected parent IDs to be {expected_parents}.")

    if feeds_rel.child_ids != ["brick:feedsAir"]:
        pytest.fail('Expected child IDs to be ["brick:feedsAir"].')

    if feeds_rel.inverse_ids != ["brick:isFedBy"]:
        pytest.fail('Expected inverse IDs to be ["brick:isFedBy"].')


def test_get_trait_by_id() -> None:
    """Test retrieving a trait by ID from a domain using get_trait_by_id."""
    domains = _validate_domains(DOMAINS_RESPONSE["data"])
    if len(domains) != 16:  # noqa: PLR2004
        pytest.fail("Expected 16 domains, found a different number.")

    if domains[2].name != "Brick":
        pytest.fail('Expected the third domain to have name "Brick".')

    brick_domain = domains[2]
    retrieved_trait = brick_domain.get_trait_by_id("brick:Building")
    if retrieved_trait.name != "Building":
        pytest.fail('Expected the retrieved trait name to be "Building".')


def test_get_parent_traits() -> None:
    """Test retrieving parent traits from a domain using get_parent_traits."""
    domains = _validate_domains(DOMAINS_RESPONSE["data"])
    brick_domain = domains[2]
    trait = brick_domain.get_trait_by_id("brick:Air_Temperature_Sensor")
    parent_traits = get_parent_traits(trait, domains)

    expected_ids = {
        "brick:Air_Temperature_Sensor",
        "brick:Point",
        "brick:Sensor",
        "brick:Temperature_Sensor",
    }
    actual_ids = {t.id for t in parent_traits}
    if actual_ids != expected_ids:
        pytest.fail(f"Expected parent trait IDs {expected_ids}, got {actual_ids}.")


def test_get_trait_by_id_exceptions() -> None:
    """Test error handling for get_trait_by_id with missing traits or empty domains."""
    domains = _validate_domains(DOMAINS_RESPONSE["data"])
    brick_domain = domains[2]

    # Test with non-existent ID
    with pytest.raises(KeyError, match=r".*non:existent:id.*"):
        brick_domain.get_trait_by_id("non:existent:id")

    # Test with empty trait list
    empty_data = {
        "domainDefinitions": [
            {
                "name": "Empty",
                "prefix": "empty",
                "description": "Empty domain",
                "uri": "http://example.org/empty",
                "traits": None,
                "relationships": None,
            },
        ],
    }
    empty_domain = _validate_domains(empty_data)[0]
    with pytest.raises(KeyError, match=r".*any:id.*"):
        empty_domain.get_trait_by_id("any:id")


def test_get_parent_traits_exceptions() -> None:
    """Error handling for get_parent_traits, non-existent/cyclic references."""
    domains = _validate_domains(DOMAINS_RESPONSE["data"])
    brick_domain = domains[2]

    # Test with non-existent parent IDs
    trait_no_parents = brick_domain.get_trait_by_id("brick:Point")
    parent_traits = get_parent_traits(trait_no_parents, domains)
    if len(parent_traits) != 1:
        pytest.fail("Expected 1 parent trait, found a different number.")
    if parent_traits[0].id != "brick:Point":
        pytest.fail('Expected parent trait ID to be "brick:Point".')

    # Test with cyclic parent references
    cyclic_data = {
        "domainDefinitions": [
            {
                "name": "Cyclic",
                "prefix": "cyclic",
                "description": "Cyclic domain",
                "uri": "http://example.org/cyclic",
                "traits": [
                    {
                        "name": "Trait1",
                        "id": "cyclic:trait1",
                        "parent_ids": [{"id": "cyclic:trait2"}],
                        "child_ids": [],
                        "equivalent_ids": [],
                        "domain_prefix": {"prefix": "cyclic"},
                    },
                    {
                        "name": "Trait2",
                        "id": "cyclic:trait2",
                        "parent_ids": [{"id": "cyclic:trait1"}],
                        "child_ids": [],
                        "equivalent_ids": [],
                        "domain_prefix": {"prefix": "cyclic"},
                    },
                ],
                "relationships": None,
            },
        ],
    }
    cyclic_domain = _validate_domains(cyclic_data)[0]
    with pytest.raises(RecursionError):
        get_parent_traits(
            cyclic_domain.get_trait_by_id("cyclic:trait1"),
            [cyclic_domain],
        )

    # Test with None parent_ids
    trait_dict = {
        "name": "NoParents",
        "id": "test:no_parents",
        "parent_ids": None,
        "child_ids": [],
        "equivalent_ids": [],
        "domain_prefix": {"prefix": "test"},
    }
    no_parent_trait = Trait.model_validate(trait_dict)
    parent_traits = get_parent_traits(no_parent_trait, brick_domain)
    if len(parent_traits) != 1:
        pytest.fail(
            "Expected 1 parent trait with None parent_ids, found a different number.",
        )
    if parent_traits[0].id != "test:no_parents":
        pytest.fail('Expected the trait ID to be "test:no_parents".')


def test_get_trait_by_id_with_domains() -> None:
    """Test retrieving a trait by ID from a list of domains using get_trait_by_id."""
    domains = _validate_domains(DOMAINS_RESPONSE["data"])
    trait = get_trait_by_id("brick:Building", domains)
    if trait.name != "Building":
        pytest.fail('Expected the retrieved trait name to be "Building".')

    # Test with non-existent ID across domains
    with pytest.raises(ValueError, match=r".*nonexistent:id.*"):
        get_trait_by_id("nonexistent:id", domains)

    # Test with empty domain list
    with pytest.raises(ValueError, match=r".*any:id.*"):
        get_trait_by_id("any:id", [])

    # Test with None domain list
    with pytest.raises(TypeError, match=r".*None.*"):
        get_trait_by_id("any:id", None)

    # Test with invalid domain data
    invalid_domains = [None, "not a domain", 123]
    with pytest.raises(AttributeError):
        get_trait_by_id("any:id", invalid_domains)



"""Test cases for QueryManager functionality."""

import pytest
from graphql import parse, print_ast

from piscada_foresight.queries_templates.query_manager import QueryManager
from tests.utils.test_utils import load_file_content


def test_singleton_behavior() -> None:
    """Ensure that QueryManager is a singleton.

    Create two instances with different domains and verify that they refer
    to the same instance.
    """
    qm1 = QueryManager(domain="domain1",
                       client_id=None,
                       client_secret=None,
                       graphql_endpoint=None,
                       json_serialize=None,
                       )
    qm2 = QueryManager(domain="domain2",
                       client_id=None,
                       client_secret=None,
                       graphql_endpoint=None,
                       json_serialize=None,
                       )
    if qm1 is not qm2:
        pytest.fail(
            "QueryManager should return the same instance regardless of "
            "constructor args.",
        )


def test_load_queries_build_dict() -> None:
    """Test that _load_queries populates the _query_dict.

    The query dictionary should match the expected values and contain
    required keys.
    """
    qm = QueryManager(domain="fake_domain",
                      client_id=None,
                      client_secret=None,
                      graphql_endpoint=None,
                      json_serialize=None,
                      )
    qdict = qm.query_handler.query_dict

    expected_dict = {
        "get_aggregated_values": "graphql_queries/timeseries/"
                                "get_aggregated_values.j2",
        "get_domains": "graphql_queries/domains/get_domains.graphql",
        "get_latest_value": "graphql_queries/timeseries/"
                            "get_latest_value.j2",
        "get_raw_values": "graphql_queries/timeseries/"
                          "get_raw_values.j2",
    }
    if qdict != expected_dict:
        pytest.fail("The query dictionary does not match the expected values.")

    for key in expected_dict:
        if key not in qdict:
            pytest.fail(f"Key '{key}' is missing from the query dictionary.")


def test_load_query_not_found() -> None:
    """If a non-existent query is requested, a ValueError should be raised.

    This ensures that queries missing from _query_dict trigger an exception.
    """
    qm = QueryManager(domain="fake_domain",
                      client_id=None,
                      client_secret=None,
                      graphql_endpoint=None,
                      json_serialize=None,
                      )
    with pytest.raises(
        ValueError,
        match=("Query 'does_not_exist' not found in query_dict"),
    ):
        qm.load_query("does_not_exist", {})


def test_load_query_j2() -> None:
    """Test loading a Jinja (.j2) query file.

    Ensure that the final query text is correctly rendered from the Jinja
    environment.
    """
    qm = QueryManager(domain="fake_domain",
                      client_id=None,
                      client_secret=None,
                      graphql_endpoint=None,
                      json_serialize=None,
                      )
    entity_ids = [0, 1]
    entity_variables = {
        f"entityId_{i}": entity_id
        for i, entity_id in enumerate(entity_ids)
    }
    jinja_dict = {"variable_names": list(entity_variables.keys())}
    rendered_text = qm.load_query("get_raw_values", jinja_dict)
    expected_text = load_file_content("assert_results/rendered_jinja_raw.graphql")

    rendered_ast = parse(rendered_text)
    expected_ast = parse(expected_text)

    rendered_normalized = print_ast(rendered_ast)
    expected_normalized = print_ast(expected_ast)

    if rendered_normalized != expected_normalized:
        pytest.fail("The rendered GraphQL query does not match the expected output.")


def test_load_query_graphql() -> None:
    """Loading a .graphql returns its contents directly (no rendering).

    This confirms that GraphQL queries are loaded as raw text.

    """
    qm = QueryManager(domain="fake_domain",
                      client_id=None,
                      client_secret=None,
                      graphql_endpoint=None,
                      json_serialize=None,
                      )
    query_text = qm.load_query("get_domains", {})
    expected_query = load_file_content(
        "../../src/piscada_foresight/queries_templates/"
        "graphql_queries/domains/get_domains.graphql",
    )
    if query_text != expected_query:
        pytest.fail("The loaded GraphQL query does not match the expected content.")
    if qm.query_handler.query_text != query_text:
        pytest.fail("QueryManager.query_text should store the loaded query text.")



def test_load_query_unsupported_extension() -> None:
    """Test ValueError is raised when loading an unsupported file.

    This is simulated by manually inserting an unsupported extension in the
    query dictionary.
    """
    qm = QueryManager(domain="fake_domain",
                      client_id=None,
                      client_secret=None,
                      graphql_endpoint=None,
                      json_serialize=None,
                      )
    qdict = qm.query_handler.query_dict
    qdict["bad_query"] = "folder/bad_query.txt"
    with pytest.raises(
        ValueError,
        match=("Unsupported query file type: 'folder/bad_query.txt'"),
    ):
        qm.load_query("bad_query", [])



"""Authenticated HTTPXTransport used to connect to Foresight.

This module provides a transport class for connecting to Foresight's GraphQL API using
OAuth2 authentication. It handles token management, automatic refresh, and interactive
browser-based login when needed.

The main class is ForesightHTTPXTransport which can be used with the gql library
to make authenticated GraphQL requests.
"""

import http.server
import json
import logging
import socket
import socketserver
import webbrowser
from contextlib import closing
from pathlib import Path
from typing import Any, Callable

import httpx
from authlib.common.security import generate_token
from authlib.integrations.httpx_client import OAuth2Client, OAuthError
from authlib.oauth2.rfc7636 import create_s256_code_challenge
from gql.transport.exceptions import TransportAlreadyConnected
from gql.transport.httpx import HTTPXTransport

SCOPE = "email"
CODE_CHALLENGE_METHOD = "S256"

log = logging.getLogger(__name__)


def _find_free_port() -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def _get_config(domain: str) -> dict:
    config_response = httpx.get(
        f"https://accounts.{domain}/realms/foresight/.well-known/openid-configuration",
        timeout=10.0,
    )
    if config_response.status_code != httpx.codes.OK:
        msg = f"Could not fetch authentication configuration from {domain}."
        raise RuntimeError(
            msg,
        )
    return config_response.json()


def _get_state_file(client_id: str) -> Path:
    return Path().home() / Path(f".{client_id}_state")


def _has_state_and_refresh_token(domain: str, client_id: str) -> bool:
    try:
        persisted_state = json.load(_get_state_file(client_id).open())
    except FileNotFoundError:
        return False
    return (
        domain in persisted_state
        and "state" in persisted_state[domain]
        and "refresh_token" in persisted_state[domain]
    )


def _get_state_and_refresh_token(domain: str, client_id: str) -> tuple[str, str]:
    persisted_state = json.load(_get_state_file(client_id).open())
    state = persisted_state[domain]["state"]
    refresh_token = persisted_state[domain]["refresh_token"]
    return (state, refresh_token)


def _save_state_and_refresh_token(
    domain: str, client_id: str, state: str, refresh_token: str,
) -> None:
    json.dump(
        {domain: {"state": state, "refresh_token": refresh_token}},
        _get_state_file(client_id).open("w"),
    )


def _get_interactive_client(domain: str,
                            client_id: str,
                            **kwargs: object,
                            ) -> OAuth2Client:
    port = _find_free_port()
    openid_configuration = _get_config(domain)
    if _has_state_and_refresh_token(domain, client_id):
        state, refresh_token = _get_state_and_refresh_token(domain, client_id)
        client = OAuth2Client(
            client_id,
            scope=SCOPE,
            code_challenge_method=CODE_CHALLENGE_METHOD,
            state=state,
            **kwargs,
        )
        try:
            token = client.refresh_token(
                openid_configuration["token_endpoint"], refresh_token=refresh_token,
            )
            _save_state_and_refresh_token(
                domain, client_id, state, token["refresh_token"],
            )
        except OAuthError:
            _get_state_file(client_id).unlink(missing_ok=True)
            return _get_interactive_client(domain, client_id, **kwargs)
        else:
            return client
    # We need to fetch a new token from scratch.
    code_verifier = generate_token(48)
    code_challenge = create_s256_code_challenge(code_verifier)

    client = OAuth2Client(
        client_id=client_id,
        redirect_uri=f"http://localhost:{port}/callback",
        code_verifier=code_verifier,
        **kwargs,
    )

    uri, state = client.create_authorization_url(
        openid_configuration["authorization_endpoint"],
        scope=SCOPE,
        code_challenge=code_challenge,
        code_challenge_method="S256",
    )

    class CallbackHandler(http.server.SimpleHTTPRequestHandler):
        def do_GET(self) -> None: # noqa: N802
            if self.path.startswith("/callback"):
                authorization_response = self.path

                token = client.fetch_token(
                    openid_configuration["token_endpoint"],
                    authorization_response=authorization_response,
                    code_verifier=code_verifier,
                )
                _save_state_and_refresh_token(
                    domain, client_id, state, token["refresh_token"],
                )

                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(b"Authorization complete. You can close this window.")
            else:
                self.send_error(404)
                msg = "Could not retrieve response."
                raise RuntimeError(msg)

    with socketserver.TCPServer(("localhost", port), CallbackHandler) as httpd:
        webbrowser.open(uri)
        httpd.handle_request()

    return client


def _get_non_interactive_client(
    domain: str, client_id: str, client_secret: str, **kwargs: object,
) -> OAuth2Client:
    """Create an OAuth2Client using client credentials flow.

    Args:
        domain: The Foresight domain to authenticate against
        client_id: OAuth2 client ID
        client_secret: OAuth2 client secret
        **kwargs: Additional arguments passed to OAuth2Client

    Returns:
        OAuth2Client configured with access token

    """
    openid_configuration = _get_config(domain)

    client = OAuth2Client(
        client_id=client_id, client_secret=client_secret, scope=SCOPE, **kwargs,
    )

    client.fetch_token(
        openid_configuration["token_endpoint"], grant_type="client_credentials",
    )

    return client


class ForesightHTTPXTransport(HTTPXTransport):
    """Sync HTTPXTransport using OAuth2 authentication with Foresight.

    This transport handles OAuth2 authentication flow with Foresight, including
    automatic token refresh and browser-based login when needed. It manages state
    and refresh tokens persistently between sessions.

    Examples:
        transport = ForesightHTTPXTransport(
            domain="foresight.example.com",
            client_id="my-client"
        )
        client = Client(transport=transport)

    """

    def __init__(
            self,
            domain: str,
            client_id: str | None = "foresight-lib-py",
            client_secret: str | None = None,
            graphql_endpoint: str | None = None,
            json_serialize: Callable[[Any], str] = json.dumps,
            **kwargs: object,
    ) -> None:
        """Create a new ForesightHTTPXTransport.

        Uses 'Authorization Code Flow with Proof Key for Code Exchange (PKCE)',
        when only client_id is given (default). Set client_id and client_secret to use
        Client Credentials Flow.
        Put both client_id and client_secret (=None) to disable auth.

        Parameters
        ----------
        domain : str
            Which Foresight domain to authenticate against.
        client_id : str, optional
            The OAuth2 client ID to use for authentication. Defaults:"foresight-lib-py".
        client_secret : str, optional
            The OAuth2 client secret to use for authentication. Defaults to "None".
        graphql_endpoint : str, optional
            The GraphQL endpoint URL. If not provided, defaults to "https://graphql.{domain}/".
        json_serialize : callable, optional
            Function to serialize JSON data, by default json.dumps.
        **kwargs : dict
            Additional arguments passed to the underlying OAuth2Client.

        """
        self.domain = domain
        self.client_id = client_id
        self.client_secret = client_secret
        if graphql_endpoint:
            super().__init__(graphql_endpoint, json_serialize, **kwargs)
        else:
            super().__init__(f"https://graphql.{domain}/", json_serialize, **kwargs)

    def connect(self) -> None:
        """Instantiate a new authenticated client connection.

        This method handles the OAuth2 authentication flow, including automatic
        token refresh and browser-based login when needed.

        Raises
        ------
        TransportAlreadyConnected
            If a client connection has already been established.
        RuntimeError
            If authentication fails or the response cannot be retrieved.

        """
        if self.client:
            msg = "Transport is already connected"
            raise TransportAlreadyConnected(msg)

        log.debug("Connecting transport")

        if self.client_id and not self.client_secret:
            self.client = _get_interactive_client(
                self.domain, self.client_id, **self.kwargs,
            )
        elif self.client_id and self.client_secret:
            self.client = _get_non_interactive_client(
                self.domain, self.client_id, self.client_secret, **self.kwargs,
            )
        elif not self.client_id and not self.client_secret:
            self.client = httpx.Client(**self.kwargs)



"""Manages queries and GraphQL client."""

from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import Any, Callable, ClassVar, Self

from gql import Client, gql

from piscada_foresight.http_piscada import ForesightHTTPXTransport
from piscada_foresight.queries_templates.query_handler import QueryHandler

log = logging.getLogger(__name__)


class QueryManager:
    """Manages queries and GraphQL client.

    This class implements a singleton pattern to maintain a single
    set of loaded queries and allow consistent access to an HTTPX-based GraphQL client.
    """

    _instance: ClassVar[QueryManager | None] = None

    def __new__(cls, *args: object, **kwargs: object) -> Self|QueryManager:
        """Create or return the existing singleton instance of QueryManager."""
        _, _ = args, kwargs
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(
        self,
        domain: str,
        client_id: str | None = None,
        client_secret: str | None = None,
        graphql_endpoint: str | None = None,
        json_serialize: Callable[[Any], str] | None = None,
        **kwargs: object,
    ) -> None:
        """Initialize QueryManager to connect to the GraphQL endpoint.

        Args:
            domain (str): The Piscada Foresight domain.
            client_id (str): The OAuth2 client identifier used for authentication.
            client_secret (str): The OAuth2 client secret used for authentication.
            graphql_endpoint (str): The URL of the Piscada Foresight GraphQL endpoint.
            json_serialize (Callable[[Any], str]): Serializes object into JSON string.
            **kwargs (object): Additional keyword argument.

        """
        self.domain = domain
        self.client_id = client_id
        self.client_secret = client_secret
        self.graphql_endpoint = graphql_endpoint
        self.json_serialize = json_serialize
        self.kwargs = kwargs
        self.transport = self.create_transport()
        self.client: Client = self.create_client()
        # Instantiate QueryHandler to manage query logic.
        self.query_handler = QueryHandler(base_directory=str(Path(__file__).parent))
        super().__init__()

    def create_transport(self) -> ForesightHTTPXTransport:
        """Create and return a ForesightHTTPXTransport."""
        json_serialize = self.json_serialize if \
            self.json_serialize is not None else json.dumps
        return ForesightHTTPXTransport(
            self.domain,
            self.client_id,
            self.client_secret,
            self.graphql_endpoint,
            json_serialize,
            **self.kwargs,
        )

    def create_client(self) -> Client:
        """Create and return a GraphQL client using the current transport."""
        return Client(transport=self.transport, fetch_schema_from_transport=False)

    def update_transport(
        self,
        client_id: str | None = None,
        client_secret: str | None = None,
        graphql_endpoint: str | None = None,
        json_serialize: Callable[[Any], str] | None = None,
        **kwargs: object,
    ) -> ForesightHTTPXTransport:
        """Update the transport configuration."""
        if client_id is not None:
            self.client_id = client_id
        if client_secret is not None:
            self.client_secret = client_secret
        if graphql_endpoint is not None:
            self.graphql_endpoint = graphql_endpoint
        if json_serialize is not None:
            self.json_serialize = json_serialize
        self.kwargs.update(kwargs)

        self.transport = self.create_transport()
        self.client.transport = self.transport
        return self.transport

    def load_queries(self) -> dict[str, str]:
        """Load query templates using QueryHandler."""
        return self.query_handler.load_queries()

    def add_queries_from_directory(self, directory: str) -> None:
        """Delegate adding queries from a new directory to QueryHandler."""
        self.query_handler.add_queries_from_directory(directory)

    def load_query(self,
                   query_name: str,
                   jinja_variables: dict[str, Any] | None,
                   ) -> str:
        """Load and render a query using QueryHandler."""
        return self.query_handler.load_query(query_name, jinja_variables)

    def add_query(self, query_name: str, query_path: str) -> None:
        """Add a new query using QueryHandler."""
        self.query_handler.add_query(query_name, query_path)

    def execute_query(self, query: str, variables: dict[str, Any]) -> dict[str, Any]:
        """Execute a GraphQL query with given variables.

        Args:
            query (str): The query string to execute.
            variables (dict[str, Any]): Variables for the query.

        Returns:
            A dictionary representing the GraphQL response.

        """
        try:
            response = self.client.execute(gql(query), variable_values=variables)
        except Exception as execution_error:
            msg = (
                f"An error occurred while executing the query: {execution_error}.\n"
                f"Query: {query}\nVariables: {variables}"
            )
            raise RuntimeError(msg) from execution_error
        return response

    def run_query(
        self,
        query_name: str,
        jinja_variables: dict[str, Any] | None = None,
        query_variables: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Load and execute a query.

        Args:
            query_name (str): The query's name.
            jinja_variables (dict[str, Any] | None): for Jinja template rendering.
            query_variables (dict[str, Any] | None): for the GraphQL query.

        Returns:
            The GraphQL response as a dictionary.

        """
        if jinja_variables is None:
            jinja_variables = {}
        if query_variables is None:
            query_variables = {}

        query_text = self.load_query(query_name, jinja_variables)
        return self.execute_query(query_text, query_variables)



"""Functions for working with domains and traits in Piscada Foresight."""
from typing import Any

from piscada_foresight.model import Domain, Trait
from piscada_foresight.queries_templates.query_manager import (
    QueryManager,
)


def get_domains(query_manager: QueryManager) -> list[Domain]:
    """Retrieve a list of all available domains.

    Parameters
    ----------
    query_manager : QueryManager
        Execcute queries against a Piscada Foresight GraphQL endpoint.

    Returns
    -------
    list[Domain]
        All available domains.

    """
    response = query_manager.run_query("get_domains")
    return _validate_domains(response)


def _validate_domains(response: dict[str, Any]) ->list[Domain]:
    """Validate and parse domain definitions from GraphQL response.

    Parameters
    ----------
    response : dict
        GraphQL response containing domain definitions

    Returns
    -------
    list[Domain]
        List of validated Domain objects

    """
    return [Domain.model_validate(domain_dict) \
               for domain_dict in response["domainDefinitions"]]


def get_parent_traits(
    trait: Trait, domains: list[Domain], *, include_self: bool = True,
) -> list[Trait]:
    """Get all parent traits for a given trait recursively.

    Traverses up the trait hierarchy to collect all parent traits, optionally including
    the input trait itself.

    Parameters
    ----------
    trait : Trait
        The trait to get parents for.
    domains : list[Domain]
        List of domains containing the trait hierarchy.
    include_self : bool
        Whether to include the input trait in the result.

    Returns
    -------
    list[Trait]
        List of parent traits, ordered from most specific to most general.

    """
    parents = []
    if include_self:
        parents.append(trait)

    if trait.parent_ids:
        for parent_id in trait.parent_ids:
            parent = get_trait_by_id(parent_id, domains)
            parents.append(parent)
            if parent.parent_ids:
                parents.extend(get_parent_traits(parent, domains, include_self=False))

    return parents


def get_child_traits(
    trait: Trait, domain: Domain, *, include_self: bool = True,
) -> list[Trait]:
    """Get all child traits for a given trait recursively.

    Traverses down the trait hierarchy to collect all child traits, optionally including
    the input trait itself.

    Parameters
    ----------
    trait : Trait
        The trait to get children for.
    domain : Domain
        The domain containing the trait hierarchy.
    include_self : bool
        Whether to include the input trait in the result.

    Returns
    -------
    list[Trait]
        List of child traits, ordered from most general to most specific.

    """
    children = []
    if include_self:
        children.append(trait)

    if trait.child_ids:
        children.extend(
            [domain.get_trait_by_id(child_id) for child_id in trait.child_ids],
        )
        for child in children[1:]:
            if child.child_ids:
                children.extend(get_child_traits(child, domain, include_self=False))

    return children


def get_trait_by_id(trait_id: str, domains: list[Domain]) -> Trait:
    """Get a trait by its ID across all domains.

    Searches through all provided domains to find a trait matching the given ID.
    The trait ID must be in the format "domain:trait_name".

    Parameters
    ----------
    trait_id : str
        ID of the trait to find in "domain:trait_name" format
    domains : list[Domain]
        List of domains to search through

    Returns
    -------
    Trait
        The found trait

    Raises
    ------
    ValueError
        If the trait ID is not found in any domain or trait ID format is invalid

    """
    try:
        domain_prefix = trait_id.split(":")[0]
    except IndexError as err:
        msg = f"Invalid trait ID format: {trait_id}"
        raise ValueError(msg) from err

    for domain in domains:
        if domain.prefix == domain_prefix:
            try:
                return domain.get_trait_by_id(trait_id)
            except KeyError:
                break

    msg = f"Trait with ID '{trait_id}' not found in domain '{domain_prefix}'"
    raise ValueError(
        msg,
    )



"""Tests for the TimeseriesResponseParser functionality.

This module tests both raw and aggregated responses, covering cases for missing
keys, empty responses, and handling of sublists.
"""

import logging
from datetime import datetime, timezone

import pytest
from pandas import DataFrame

from piscada_foresight.parser_graphql_to_df. \
    time_series.timeseries_response_parser import (
    TimeseriesResponseParser,
)


def test_get_values_aggregated_missing_values(
    mock_query_manager: object,
    aggregated_no_values: object,
    caplog: object,
) -> None:
    """Checks behavior when no aggregator data is available (empty values arrays).

    Ensure a warning is logged and an empty DataFrame is returned.
    """
    mock_query_manager.run_query.return_value = aggregated_no_values

    parser = TimeseriesResponseParser()

    entity_variables = {
        f"entityId_{i}": entity_id
        for i, entity_id in enumerate(["entityId_0", "entityId_1"])
    }

    with caplog.at_level("WARNING"):
        result = parser.parse(
            response=aggregated_no_values,
            entity_variables=entity_variables,
            start="2023-01-01",
            end="2023-01-02",
            query_type="aggregated",
        )
        # Check warning is logged
        assert "No data found for the requested time range" in caplog.text  # noqa: S101
        # Check result is an empty DataFrame
        assert isinstance(result, DataFrame)  # noqa: S101
        assert result.empty  # noqa: S101


@pytest.fixture
def parser() -> TimeseriesResponseParser:
    """Fixture to create a TimeseriesResponseParser instance."""
    return TimeseriesResponseParser()


def test_parse_raw_values_success(
    parser: TimeseriesResponseParser,
    raw_timeseries_response: object,
) -> None:
    """Test parsing a valid 'raw' timeseries response.

    The response includes multiple eventTimes and numeric values.
    """
    entity_variables = {
        "entityId_0": (
            "brick:Supply_Air_Temperature_Sensor:00000000-0000-0000-0000-"
            "000000000001"
        ),
        "entityId_1": (
            "brick:Effective_Supply_Air_Temperature_Setpoint:00000000-0000-0000-0000-"
            "000000000002"
        ),
    }

    dataframe = parser.parse(
        response=raw_timeseries_response,
        entity_variables=entity_variables,
        start=datetime(2025, 1, 14, 9, 0, 0, tzinfo=timezone.utc),
        end=datetime(2025, 1, 14, 11, 0, 0, tzinfo=timezone.utc),
        query_type="raw",
    )

    # We expect two columns, one per entity name
    assert isinstance(dataframe, DataFrame)  # noqa: S101
    assert not dataframe.empty, "DataFrame should not be empty for valid raw data"  # noqa: S101
    assert "360001 Temperature" in dataframe.columns  # noqa: S101
    assert "360001 Temperature Setpoint" in dataframe.columns  # noqa: S101

    # Check some expected values.
    # (The parser forward-fills missing times by default,
    # so you might see more rows than
    # raw data.)
    first_setpoint = dataframe["360001 Temperature Setpoint"].iloc[0]
    expected_first_setpoint = 19.5
    assert first_setpoint == expected_first_setpoint, \
        "Incorrect setpoint value in first row" # noqa: S101


def test_parse_raw_values_missing_keys(
    parser: TimeseriesResponseParser,
    raw_missing_values_tag_response: object,
) -> None:
    """Ensures RuntimeError is raised if keys are missing from response.

    For example, if the 'quantity' or 'values' keys are missing.
    """
    entity_variables = {"entityId_0": "fake_entity_id"}

    with pytest.raises(
        RuntimeError,
        match="Could not retrieve raw values for entity 'fake_entity_id'",
    ):
        parser.parse(
            response=raw_missing_values_tag_response,
            entity_variables=entity_variables,
            start=datetime(2025, 1, 14, tzinfo=timezone.utc),
            end=datetime(2025, 1, 15, tzinfo=timezone.utc),
            query_type="raw",
        )


def test_parse_raw_values_entity_none(
    parser: TimeseriesResponseParser,
) -> None:
    """Test handling when response[variable_name] is None.

    If the response is None, a TypeError will occur when accessing subkeys,
    which should be turned into a RuntimeError by the parser.
    """
    response = {"entityId_0": None}  # This will cause a TypeError
    entity_variables = {"entityId_0": "fake_entity_id"}

    with pytest.raises(RuntimeError, match="Could not find entity 'fake_entity_id'"):
        parser.parse(
            response=response,
            entity_variables=entity_variables,
            start=datetime(2025, 1, 14, tzinfo=timezone.utc),
            end=datetime(2025, 1, 15, tzinfo=timezone.utc),
            query_type="raw",
        )


def test_parse_raw_values_no_entities(
    parser: TimeseriesResponseParser,
    caplog: object,
) -> None:
    """Test behavior when no entities or data are present.

    If the response is empty or if entity_variables is empty, an empty
    DataFrame is expected.
    """
    response = {}
    entity_variables = {}

    with caplog.at_level(logging.WARNING):
        dataframe = parser.parse(
            response=response,
            entity_variables=entity_variables,
            start=datetime(2025, 1, 14, tzinfo=timezone.utc),
            end=datetime(2025, 1, 15, tzinfo=timezone.utc),
            query_type="raw",
        )
    assert dataframe.empty, \
        "Should return an empty DataFrame when no entities or data are present."  # noqa: S101


def test_parse_aggregated_values_success(
    parser: TimeseriesResponseParser,
    aggregated_timeseries_response: object,
) -> None:
    """Test parsing a valid 'aggregated' timeseries response.

    The response includes several aggregationFunctions for each entity.
    """
    entity_variables = {
        "entityId_0": (
            "brick:Supply_Air_Temperature_Sensor:0000-0000-0000-0000-"
            "000000000001"
        ),
        "entityId_1": (
            "brick:Effective_Supply_Air_Temperature_Setpoint:0000-0000-0000-0000-"
            "000000000002"
        ),
    }

    dataframe = parser.parse(
        response=aggregated_timeseries_response,
        entity_variables=entity_variables,
        start=datetime(2025, 1, 14, 9, tzinfo=timezone.utc),
        end=datetime(2025, 1, 14, 11, tzinfo=timezone.utc),
        query_type="aggregated",
    )

    # We expect columns like "360001 Temperature|min", "360001 Temperature|max", etc.
    expected_cols_entity0 = [
        "360001 Temperature|min",
        "360001 Temperature|max",
        "360001 Temperature|avg",
        "360001 Temperature|count",
        "360001 Temperature|last",
    ]
    for col in expected_cols_entity0:
        assert col in dataframe.columns  # noqa: S101

    expected_cols_entity1 = [
        "360002 Temperature Setpoint|min",
        "360002 Temperature Setpoint|max",
        "360002 Temperature Setpoint|avg",
        "360002 Temperature Setpoint|count",
        "360002 Temperature Setpoint|last",
    ]
    for col in expected_cols_entity1:
        assert col in dataframe.columns  # noqa: S101

    # Check some values
    expected_value = 19.5
    assert dataframe.loc["2025-01-14 09:59:59.786000+00:00", \
        "360001 Temperature|min"] == expected_value  # noqa: S101
    expected_value = 19.8
    assert dataframe.loc["2025-01-14 09:59:59.786000+00:00", \
        "360002 Temperature Setpoint|last"] == expected_value  # noqa: S101


def test_parse_aggregated_values_missing_key(
    parser: TimeseriesResponseParser,
    aggregated_tag_missing: object,
) -> None:
    """RuntimeError  raised when keys in response are missing.

    For example, if aggregated values cannot be retrieved.
    """
    entity_variables = {"entityId_0": "brick:Supply_Air_Temperature_Sensor:0000-0000"}

    with pytest.raises(
        RuntimeError,
        match="Could not retrieve aggregated values for entity 'brick:"
        "Supply_Air_Temperature_Sensor:0000-0000'",
    ):
        parser.parse(
            response=aggregated_tag_missing,
            entity_variables=entity_variables,
            start=datetime(2025, 1, 14, tzinfo=timezone.utc),
            end=datetime(2025, 1, 15, tzinfo=timezone.utc),
            query_type="aggregated",
        )


def test_parse_aggregated_values_all_empty(
    parser: TimeseriesResponseParser,
    aggregated_empty_sublists: object,
    caplog: object,
) -> None:
    """Test behavior when all sublists in 'aggregatedTimeseries' are empty.

    A warning should be logged and an empty DataFrame returned.
    """
    entity_variables = {"entityId_0": "brick:Supply_Air_Temperature_Sensor:0000-0000"}

    with caplog.at_level(logging.WARNING):
        dataframe = parser.parse(
            response=aggregated_empty_sublists,
            entity_variables=entity_variables,
            start=datetime(2025, 1, 14, tzinfo=timezone.utc),
            end=datetime(2025, 1, 15, tzinfo=timezone.utc),
            query_type="aggregated",
        )

    assert dataframe.empty, \
        "Should return empty DataFrame if no aggregator sublists have data."  # noqa: S101
    assert "No data found for the requested time range" in caplog.text  # noqa: S101


def test_parse_aggregated_values_partial_empty(
    parser: TimeseriesResponseParser,
    aggregated_one_empty_sublist_response: object,
) -> None:
    """Test behavior when some aggregator sublists are empty.

    Only the non-empty columns should be returned.
    """
    entity_variables = {"entityId_0": "brick:Supply_Air_Temperature_Sensor:0000-0000"}

    dataframe = parser.parse(
        response=aggregated_one_empty_sublist_response,
        entity_variables=entity_variables,
        start=datetime(2025, 1, 14, tzinfo=timezone.utc),
        end=datetime(2025, 1, 15, tzinfo=timezone.utc),
        query_type="aggregated",
    )
    assert not dataframe.empty, \
        "Should not be empty if at least one aggregator sublist has data"  # noqa: S101
    assert "360001 Temperature|max" in dataframe.columns  # noqa: S101
    assert "360001 Temperature|min" not in dataframe.columns, \
        "Empty aggregator sublist shouldn't appear as a column"  # noqa: S101
    expected_value =  19.7
    assert dataframe.iloc[0, 0] == expected_value, "Unexpected aggregated value"  # noqa: S101


def test_parse_unknown_query_type(
    parser: TimeseriesResponseParser,
) -> None:
    """Test that an unknown query_type raises a ValueError."""
    with pytest.raises(ValueError, match="Unknown query_type: invalid_type"):
        parser.parse(
            response={},
            entity_variables={},
            start=datetime(2025, 1, 14, tzinfo=timezone.utc),
            end=datetime(2025, 1, 15, tzinfo=timezone.utc),
            query_type="invalid_type",
        )



"""Test the function get_state_file."""

import pytest

from piscada_foresight import http_piscada


def test_get_state_file() -> None:
    """Test retrieving the state file path.

    This test ensures that the state file path ends with the correct basename
    (/.test_state) and starts with a slash.
    """
    state_file = http_piscada._get_state_file("test")  # noqa: SLF001
    if not str(state_file).endswith("/.test_state"):
        pytest.fail("State file path does not end with '/.test_state'")
    if not str(state_file).startswith("/"):
        pytest.fail("State file path does not start with '/'")



"""Define models for domains, traits, and relationships.

Used in the Piscada Foresight project.
"""

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, field_validator


class Relationship(BaseModel):
    """Model representing a relationship between entities in a domain.

    Contains information about the hierarchy of relationships (parents/children)
    and inverse relationships.
    """

    name: str
    id: str
    parent_ids: List[str]
    child_ids: List[str]
    inverse_ids: List[str]
    domain_prefix: str

    @field_validator("domain_prefix", mode="before")
    @classmethod
    def extract_domain_prefix(cls, value: Dict[str, str]) -> Optional[str]:
        """Extract the domain prefix from a dict containing a 'prefix' key."""
        return value["prefix"] if value else None

    @field_validator("parent_ids", mode="before")
    @classmethod
    def extract_parent_ids(cls, value: List[Dict[str, str]]) -> List[str]:
        """Extract a list of parent ids from a list of dictionaries."""
        return [e["id"] for e in value] if value else []

    @field_validator("child_ids", mode="before")
    @classmethod
    def extract_child_ids(cls, value: List[Dict[str, str]]) -> List[str]:
        """Extract a list of child ids from a list of dictionaries."""
        return [e["id"] for e in value] if value else []

    @field_validator("inverse_ids", mode="before")
    @classmethod
    def extract_inverse_ids(cls, value: Dict[str, str]) -> List[str]:
        """Extract the inverse id from a dict and return it as a list."""
        return [value["id"]] if value else []


class Trait(BaseModel):
    """Model representing a trait that can be assigned to entities.

    Contains hierarchical information about parent/child relationships and
    equivalent traits.
    """

    name: str
    id: str
    domain_prefix: str
    parent_ids: List[str]
    child_ids: List[str]
    equivalent_ids: List[str]

    @field_validator("domain_prefix", mode="before")
    @classmethod
    def extract_domain_prefix(cls, value: Dict[str, str]) -> str|None:
        """Extract the domain prefix from a dict containing a 'prefix' key."""
        return value["prefix"] if value else None

    @field_validator("parent_ids", mode="before")
    @classmethod
    def extract_parent_ids(cls, value: list[dict[str, str]]) -> list[str]:
        """Extract a list of parent ids from a list of dictionaries."""
        return [e["id"] for e in value] if value else []

    @field_validator("child_ids", mode="before")
    @classmethod
    def extract_child_ids(cls, value: list[dict[str, str]]) -> list[str]:
        """Extract a list of child ids from a list of dictionaries."""
        return [e["id"] for e in value] if value else []

    @field_validator("equivalent_ids", mode="before")
    @classmethod
    def extract_equivalent_ids(cls, value: list[dict[str, str]]) -> list[str]:
        """Extract a list of equivalent trait ids from a list of dictionaries."""
        return [e["id"] for e in value] if value else []


class Domain(BaseModel):
    """Model representing a domain containing traits and relationships.

    A domain defines a namespace of traits and relationships that can be
    used to describe entities.
    """

    name: str
    prefix: str
    description: Optional[str]
    uri: str
    traits: List[Trait]
    relationships: List[Relationship]

    @field_validator("traits", mode="before")
    @classmethod
    def validate_traits(cls, value: object) -> list[Any]:
        """Ensure the traits field is valid, or return an empty list."""
        return  value if isinstance(value, list) else []

    @field_validator("relationships", mode="before")
    @classmethod
    def validate_relationships(cls, value: object) -> list[Any]:
        """Ensure the relationships field is valid, or return an empty list."""
        return  value if isinstance(value, list) else []

    def __init__(self, **data: object) -> None:
        """Initialize a Domain instance.

        Passes all keyword data to the base initializer and creates a trait
        lookup dictionary mapping each trait's id to its object.

        Parameters
        ----------
        **data : object
            Arbitrary keyword arguments.

        """
        super().__init__(**data)
        self._trait_dict = (
            {trait.id: trait for trait in self.traits} if self.traits else {}
        )

    def get_trait_by_id(self, id_str: str) -> Trait:
        """Retrieve a trait by its identifier.

        Parameters
        ----------
        id_str : str
            The unique identifier of the trait.

        Returns
        -------
        Trait
            The Trait object matching the given identifier.

        """
        return self._trait_dict[id_str]
